# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys
import pathlib

env_dir = pathlib.Path().home() / '.lemondb'
log_dir = env_dir / 'logs'

if not env_dir.exists():
    os.mkdir(env_dir)
elif env_dir.is_file():
    os.remove(str(env_dir.absolute()))
    os.mkdir(env_dir)

log_path = str((log_dir / 'lemon-logs').absolute())


import loguru; logger = loguru.logger
logger.remove()
logger.add(log_path, rotation="100 MB", compression='zip')
logger.add(
    sys.stdout, 
    format='<green>[{time:HH:mm:ss}]</green> |  <magenta>{level}</magenta>  | <lvl>{message}</lvl>', 
    level='INFO'
)
logger.opt(colors=True)


# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from lemondb.plugin import (
    BasePlugin,
    LemonPlugin
)
import pathlib
from lemondb.types import (
    Optional,
    Middleware,
    Any,
    Dict,
    Lambda,
    Iterable,
    Mapping
)
from lemondb.query import (
    SearchQuery,
    Linq
)
from lemondb.server import (
    LemonServer,
    LemonClient
)
import socket
from urllib.parse import urlparse
from lemondb.middleware import JsonMiddleware
from lemondb.document import Document
from lemondb.constants import ops
from lemondb.utils import iterate_dict
from lemondb.logger import logger
from lemondb.errors import SearchQueryError
import re



class LemonDB:
    """
    NOTE: This library is currently on a BETA / ALPHA. Not yet
    used for production since there is some bugs might occur.
    
    Release Changes: v.0.0.3:
        The new release v.0.0.3 has added new features. The new
        Socket Server & Client feature so that you can run the
        database on a VPS or any hosting server. 

    NOTE: For Server & Client used. Kindly pass keyword argument
    host_checking to automatically detect if it is client or
    server or manually pass keyword arguments if the given name
    is client or server to avoid slow performance. 

    LemonDB is a simple and lightweight document oriented database 
    written in pure Python 3 tried on version: `3.9` & `3.8`. It 
    should work on versions  <= 3.7. This class handle all operation 
    including storing document on a file.

    For Server & Client, make sure to use the lemondb:// as the scheme
    for the server. This recognized and parsed the host, port and the
    keyword argument given by the query string. 

    Based on performance, LemonDB comes in first place ahead of the 
    popular `TinyDB`, but it is not expected to replace `TinyDB`.
    

    Here are the result for the database operation that store 
    1000 random generated strings.

    LemonDB: 20.848030 / 20.85 seconds
    TinyDB: 53.912508 / 53.91 seconds

    It is actually 2x faster than the TinyDB. It can be a little bit
    faster since LemonDB support different type of JSON Serialization
    that is faster than the standard `json` library. It supports:
    
    - `simplejson (Estimated result for 1000 insert operation: 27.86 sec)`
    - `ujson (Estimated result for 1000 insert operation: 22.88 sec)`
    - `hyperjson (Estimated result for 1000 insert operation: 20.18 sec)`


    NOTE: LemonDB support table operation where you stored a data inside
    a table. You can create / get the table by calling the `table` method:
        
        >>> from lemondb import LemonDB
        >>> db = LemonDB('lemon.json')
        >>> names = db.table('name') #: Create / Get the table .
        >>> names.insert({'name': 'John Doe'})
        >>> {'name': 'John Doe'}

    Last but not the least, LemonDB support a database encryption with 
    password also known as Sidle Encryption (https://github.com/znqi/sidle). 
    By default LemonDB allows you to install the `sidle` library in order 
    to do the operation. You can access it by using the standard middleware:
    `lemondb.middleware.SidleMiddleware` that accept a positional arguments
    `password`. Also, make sure to include the `lemon.plugin.SidlePlugin`.

        >>> from lemondb import LemonDB
        >>> from lemondb.plugin impor SidlePlugin
        >>> from lemondb.middleware import SidleMiddleware
        >>> ...
        >>> db = (
        >>>     'test.json',
        >>>     middleware_cls=SidleMiddleware('password'),
        >>>     plugin_cls=SidlePlugin
        >>> ...

    Parameters:

        :param name (str):
            The name of the database. It can be a file name.

        :param plugin_cls (BasePlugin : Optional):
            The base plugin for Lemon DB. The plugin runs
            everytime the database is called or initialized.
            Default value: LemonPlugin

        :param middleware_cls (BaseMiddleware : Optional):
            The middleware for the document that handles read,
            write and delete operation on the file given.
            Default Value: JsonMiddleware


        :param document_cls (Document):
            Set the document class for creating documents.
            Default Value: Document

    Server Example:
        >>> from lemondb import LemonDB
        >>> db = LemonDB('lemondb://0.0.0.0:3000', server=True)

    Client Example:
        >>> from lemondb import LemonDB
        >>> db = LemonDB('lemondb://localhost:3000', client=True)
        >>> db.insert({'name': 'John Doe'})

    Example:

        >>> from lemondb import LemonDB
        >>> db = LemonDB('test.json')
        >>> db.insert({
        >>>         'name': 'John Doe'
        >>>     })
        >>> {'name': 'John Doe'}
        >>> ...
        >>> #: For query searching
        >>> from lemondb import Query
        >>> query = Query()
        >>> db.search(query.name == 'John Doe')
        >>> [{'name': 'John Doe'}]

    
    """

    #: The path for the database.
    db_path: pathlib.Path

    #: The default table for the database
    default_table: str = "_table"

    #: LemonCLient Instance
    #: versionAdded: 0.0.3
    client_instance: LemonClient = None

    #: LemonServer Instance
    #: versionAdded: 0.0.3
    server_instance: LemonClient = None


    def __init__(
        self,
        name: str,
        plugin_cls: Optional[BasePlugin] = None,
        middleware_cls: Optional[Middleware] = None,
        document_cls: Optional[Document] = None,
        **kwargs
    ):

        """
        Initialize Lemon DB

        Parameters:

            :param name (str):
                The name of the database. It can be a file name.

            :param plugin_cls (BasePlugin : Optional):
                The base plugin for Lemon DB. The plugin runs
                everytime the database is called or initialized.
                Default value: LemonPlugin

            :param middleware_cls (BaseMiddleware : Optional):
                The middleware for the document that handles read,
                write and delete operation on the file given.
                Default Value: JsonMiddleware


            :param document_cls (Document):
                Set the document class for creating documents.
                Default Value: Document

        Example:

            >>> from lemondb import LemonDB
            >>> db = LemonDB('test.json')
            >>> db.insert({'name': 'John Doe'})

        """

        self.name = name
        self.kwargs = kwargs
        self.db_path = pathlib.Path(self.name)
        self.repr_name = type(self).__name__
        self.plugin_cls = plugin_cls
        self.logger = logger

        self.server = self.kwargs.get('server', False)
        self.client = self.kwargs.get('client', False)

        if not plugin_cls:
            self.plugin_cls = LemonPlugin()
        else:
            self.plugin_cls = plugin_cls

        if not middleware_cls:
            self.middleware_cls = JsonMiddleware()
        else:
            self.middleware_cls = middleware_cls

        if not document_cls:
            self.document_cls = Document(
                path=self.db_path, 
                middleware_cls=self.middleware_cls
            )
        else:
            self.document_cls = document_cls

        
        if not 'table_name' in self.kwargs:
            self.kwargs.__setitem__('table_name', self.default_table)
        
        self.table_name = self.kwargs.get('table_name', self.default_table)
    

        if self.table_name:
            self.default_table = self.table_name

        if self.kwargs.get('host_checking', False):
            if self.__check_if_server_client():
                self.client = True
            elif self.__check_if_server_client == 0:
                self.server = True
        
        if self.server:
            parsed = self.__parse_url(self.name)
            db_dir = pathlib.Path().home() / '.lemondb' / 'db'
            if not db_dir.exists():
                os.mkdir(db_dir.absolute())
            
            self.name = str(
                (db_dir / 'db' ).absolute()
            )
            self.document_cls = Document(
                path=(db_dir / 'db' ).absolute(),
                middleware_cls=self.middleware_cls
            )

            self.run_plugin(plugin_cls=plugin_cls)
            self.plugin_cls._init_db()

            self.server_instance = LemonServer(
                host=(parsed['host'], parsed['port'])
            )
            self.server_instance.run(self)
        
        elif self.client:
            parsed = self.__parse_url(self.name)
            self.client_instance = LemonClient(
                host=(parsed['host'], parsed['port'])
            )

        self.run_plugin(plugin_cls=plugin_cls)

        
        if not self.db_path.exists():
            self.plugin_cls._init_db()

        if self.server:
            self.repr_name = 'LemonServer'
        elif self.client:
            self.repr_name = 'LemonClient'

        

    @logger.catch
    def table(self, name: str, **options):
        """
        The table for the database. If the given
        table name doesnt exist then create a new one.

        The table handles a sorted dictionary that contains
        the data.
        """

        options.__setitem__('table_name', name)
        
        return LemonDB(
            name=name, 
            plugin_cls=self.plugin_cls,
            middleware_cls=self.middleware_cls, 
            document_cls=self.document_cls,
            **options
        )

    @logger.catch
    def tables(self):
        """
        Get all table name and return a list.
        """
        return [k for k in self.document_cls.read().keys()]


    @logger.catch
    def items(self, table_name: Optional[str] = None, **options):
        """
        Return all items from the given table, packed on a single list
        """

        return_dict = options.get('dict', False)
        data = self.document_cls.read()
        if table_name:        
            _items = [data[x] for x in data.keys() if x == table_name]
        else:
            _items = [{k:v} for k,v in data.items()]

        if return_dict:
            for k,v in data.items():
                _items = [{k:v} for k,v in v.items()]


        return _items


    @logger.catch
    def clear(self):
        """
        Clear all item from the database including the tables and
        create a new default table name.
        """

        data = self.document_cls.read()
        data.clear()
        self.plugin_cls._init_db()

    @logger.catch
    def insert(self, item: Mapping, **options):
        """
        Insert a item to the database. The item should
        be a mapping `(dict)`.
        
        Parameter:
            :param item (Mapping):
                The item to be added to the database.

        Example:
            >>> from lemondb import LemonDB
            >>> db = LemonDB('test')
            >>> db.insert({'name': 'zenqi'})

        Retun:
            The item to be inserted.

        """
        if self.client_instance:
            self.client_instance.send(
                op='insert',
                data=item
            )
            return item
            
        
        raw_data = self.document_cls.read()
        raw = False
        if not self.db_path.exists():
            self.plugin_cls._init_db()
        
        table = options.pop('table', self.default_table)
        if table:
            _r_d = {}

            if table in raw_data.keys():
                for k,v in raw_data.items():
                    if k == table:
                        _r_d = {k:v}
                    
            else:
                _r_d = {table: {}}
            
            raw = True
            if table == self.default_table:

                item = self.document_cls._increment(
                    data=_r_d, item=item)


            else:
                item = self.__construct_table(
                    table=table, 
                    data=_r_d,
                    raw=item
                )
            
        self.document_cls.write(item, raw=raw)
        return item
    
    @logger.catch
    def insert_many(self, iterable: Iterable):
        """
        Simillar to `insert` however insert all items 
        from the given iterable / list. 
        """
        
        for i in iterable:
            if self.client_instance:
                self.client_instance.send(
                    op='insert_many',
                    data=i
                )
            else: self.insert(i)

        return iterable

    @logger.catch
    def delete(
        self, 
        query: Any, 
        all: Optional[bool] = True
    ):
        
        """
        Delete a key from a query given. The query accept
        3 types. Similar to `search`. 

        Parameter:
            query (Any):
                The query of the key to delete.

            all (Optional[Boolean]):
                (added on v0.0.2)

                Set if all existing keys/simillar value
                to be deleted. Default Value: `True`

        Examples:
            >>> query = Query()
            >>> db.delete(query.name == 'John Doe')
            >>> ...

        Return:
            The deleted item.

        """
        if self.client_instance:
            self.client_instance.send(
                op='delete',
                data=query
            )
        
        if isinstance(query, Mapping):
            self.document_cls.delete(query, all=True)
            return query

        else:
            try:
                if all:
                    data = self.search(query)
                else:
                    data = self.search(query)[0]
            
            except IndexError:
                # TODO: No result found on a search.
                return None

            self.document_cls.delete(data, all=True)
            return data

    @logger.catch
    def update(self, query: Any, item: Mapping):
        """
        ADDED: `v0.0.2`
        
        Update the data from the default given table name.
        This perform a `search` query that can be 3 types,
        and update the first result of the query.

        Parameters:
            query (Any):
                The query syntax for searching item

            item (Mapping):
                The item to be replace

        Example:

            >>> from lemondb import LemonDB, Query
            >>> db = LemonDB('test.json')
            >>> query = Query()
            >>> ...
            >>> db.insert({'name': 'John Doe', 'password': '1234'})
            >>> ...
            >>> db.update(
            >>>     query.name == 'John Doe', 
            >>>     {'password': 'newpassword'}
            >>> )
            >>> ...

        """

        _ = self.search(query)
        
        if not _:
            #: TODO: Searching failed. No result found
            raise SearchQueryError('The search query doesnt exist on the table/database')

        result = _[0]
        data = self.document_cls.read()
        for table, value in list(data.items()):
            for k,v in list(value.items()):
                if v == result:
                    data[table][k].update(item)
                    break

        self.document_cls.write(data, mode='w', raw=True)
        return item

    @logger.catch
    def search(self, query):
        """
        Search an item from the database. The query accept
        3 types. The first one is the standard `SearchQuery`,
        next is the `lambda` function and the `re` pattern.

        Parameter:
            query (Any):
                The query of the key to search.
    

        Example:
            >>> from lemondb import LemonDB, Query
            >>> db = LemonDB('test.json')
            >>> db.insert({'name': 'John Doe'})
            >>> ...
            >>> query = Query()
            >>> db.search(query.name == 'John Doe')
            >>> [{"name": "John Doe"}]
        
        Return:
            The list of possible result for the queries.

        """
        if self.client_instance:
            self.client_instance.send(
                op='search',
                data=query
            )

        items = self.items(dict=True)
        result = []
        use_lambda = False
        use_re = False
        use_sq = False

        if isinstance(query, Lambda):
            use_lambda = True

        elif isinstance(query, SearchQuery):
            use_sq = True

        else:
            use_re = True

        
        reconstructed_list = []
        for i in items:
            for k,v in i.items():
                if use_re:
                    for _, rv in iterate_dict(v):
                        if re.search(query, str(rv), re.IGNORECASE):
                            result.append(v)
                            
                else:
                    reconstructed_list.append(v)
        
        
        _query = Linq(reconstructed_list)
        if use_sq:
            op, key, item = query()
            lambda_wrapper = lambda x: ops[op](x[item], key)
            return _query.where(lambda_wrapper).to_list()

        if use_lambda:
            return _query.where(query).to_list()
        

        return result
        
    def __len__(self):
        data = self.document_cls.read()
        return len(data[self.default_table])


    def __repr__(self) -> str:
        return "{name}(table_name=\"{table}\", length={length}, table_count={table_count})".format(
            name=self.repr_name,
            table=self.default_table,
            length=len(self),
            table_count=len(self.tables()),
        )

    def __construct_table(
        self, 
        table: str, 
        data: Mapping, 
        raw: Optional[Mapping] = {},
    ):

        """
        Create a table for the given data
        """

        if not raw:
            _ = {table: {}}
        elif raw and data:
            _ = self.document_cls._increment(data, raw)
        else:
            _ = self.document_cls._increment({table: {}}, raw)

        data.update(_); return data

    
    def __check_if_server_client(self):
        """
        Check if the db name is server or client
        """

        # Match if the given db name is a server pattern
        pattern = re.compile(
            r'^(?:lemondb|http)s?://' # lemondb:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
            r'localhost|' #localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
            r'(?::\d+)?' # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE
        )

        m = re.match(pattern, self.name)
        
        if m:
            #: Try if it is client or server
            parsed = self.__parse_url(self.name)
            try:
                sock = socket.socket()
                sock.connect((parsed['host'], parsed['port']))
                return 1 #: 1 means client

            except socket.error:
                return 0 #: 0 means server
            

        return None

    def __parse_url(self, url: str):
        """
        Parse the url and return a dictionary.
        Version Added: 0.0.3
        """

        parsed = urlparse(url)
        return {
            'scheme': parsed.scheme,
            'host': parsed.hostname,
            'port': parsed.port,
            'query': parsed.query,
        }

    def run_plugin(self, plugin_cls: Any):
        """
        Seperate function to run plugin.
        Version Added: 0.0.3
        """

        try:
            #: Run the plugin and give all parameters
            self.plugin_cls.run(
                name=self.name,
                document_cls=self.document_cls,
                plugin_cls=self.plugin_cls,
                middleware_cls=self.middleware_cls,
                **self.kwargs
            )
        except TypeError:
            self.plugin_cls = plugin_cls()
            self.plugin_cls.run(
                name=self.name,
                document_cls=self.document_cls,
                plugin_cls=self.plugin_cls,
                middleware_cls=self.middleware_cls,
                **self.kwargs
            )



# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import (
    Optional,
    Any,
    IO,
    Iterable,
    Mapping,
    TextIO,
    Type,
    TypedDict,
    Dict,
    List
)
from lemondb.middleware.base import BaseMiddleware
from lemondb.serializer.base import BaseSerializer


class DocumentDict(TypedDict):
    item: dict
    ...

Lambda = type(lambda: None)
Middleware = BaseMiddleware


__all__ = [
    'Optional',
    'Any',
    'IO',
    'Iterable',
    'Mapping',
    'TextIO',
    'Type',
    'TypedDict',
    'Dict',
    'List',
    'Document',
    'BaseSerializer'
]


# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import itertools
from lemondb.types import Iterable

class SearchQuery:
    """
    A search query instance
    """
    _ops: str = None
    _key: str = None
    _item: str = None
    def __init__(self, data=None):
        self._data = data


    def __eq__(self, obj: object):
        """
        Handle == operation
        """ 
        self._ops = '=='
        self._key = obj
        return self
    
    def __lt__(self, obj: object):
        """
        Handle < operation
        """

        self._ops = "<"
        self._key = obj
        return self
    
    def __gt__(self, obj: object):
        """
        Handle > operation
        """

        self._ops = ">"
        self._key = obj
        return self

    def __ge__(self, obj: object):
        """
        Handle >= operation
        """

        self._ops = ">="
        self._key = obj
        return self

    def __le__(self, obj: object):
        """
        Handle <= operation
        """

        self._ops = "<="
        self._key = obj
        return self

    def __ne__(self, obj: object):
        """
        Handle != operation
        """
        
        self._ops = "!="
        self._key = obj
        return self

    def __repr__(self):
        return "%s(key='%s', item='%s', operator=%s)" % (
            type(self).__name__,
            self._key,
            self._item,
            self._ops
        )

    def __call__(self):
        return self._ops, self._key, self._item

class Query(SearchQuery):
    """
    pass
    """
    
    def __init__(self):
        self._data = None
        super(Query, self).__init__(self._data)

    def __getattr__(self, key: str):
        """
        Get attribute. `Query().name == 'John Doe'` 
        """
        _query = SearchQuery(self._data)
        _query._item = key 
        return _query

class Linq:
    """
    A class similar to Linq (Language Integrated Query)
    that was introduced on C# (Microsoft.NET)
    """

    def __init__(self, iterable: Iterable):
        """
        Instantiates Linq wrapper
        :param iterable: iterable to wrap
        """
  
        if iterable.__class__ is Linq:
            self.iterable = iterable.iterable
        else:
            self.iterable = iterable

    def __repr__(self):
        return repr(self.to_list())

    def __iter__(self):
        """
        Allows to iterate Linq object
        """
        return iter(self.iterable)

    def any(self, predicate=None):
        """
        Returns true if there any item which matches given predicate.
        If no predicate given returns True if there is any item at all.
        :param predicate: Function which takes item as argument and returns bool
        :returns: Boolean
        :rtype: bool
        """
        for i in self.iterable:
            if predicate is None:
                return True
            elif predicate(i):
                return True
        return False

    def all(self, predicate):
        """
        Returns true if all items match given predicate.
        :param predicate: Function which takes item as argument and returns bool
        :returns: Boolean
        :rtype: bool
        """
        for i in self.iterable:
            if not predicate(i):
                return False
        return True

    def first(self, predicate=None):
        """
        Returns first item which matches predicate or first item if no predicate given.
        Raises exception, if no matching items found.
        :param predicate: Function which takes item as argument and returns bool
        :returns: item
        :rtype: object
        """
        for i in self.iterable:
            if predicate is None:
                return i
            elif predicate(i):
                return i
        raise Linq.LinqException('No matching items!')

    def first_or_none(self, predicate=None):
        """
        Returns first item which matches predicate or first item if no predicate given.
        Returns None, if no matching items found.
        :param predicate: Function which takes item as argument and returns bool
        :returns: item
        :rtype: object
        """
        try:
            return self.first(predicate)
        except Linq.LinqException:
            return None

    def last(self, predicate=None):
        """
        Returns last item which matches predicate or last item if no predicate given.
        Raises exception, if no matching items found.
        :param predicate: Function which takes item as argument and returns bool
        :returns: item
        :rtype: object
        """
        last_item = None
        last_item_set = False
        for i in self.iterable:
            if (predicate is not None and predicate(i)) or (predicate is None):
                last_item = i
                last_item_set = True

        if not last_item_set:
            raise Linq.LinqException('No matching items!')
        return last_item

    def last_or_none(self, predicate=None):
        """
        Returns last item which matches predicate or last item if no predicate given.
        Returns None, if no matching items found.
        :param predicate: Function which takes item as argument and returns bool
        :returns: item
        :rtype: object
        """
        try:
            return self.last(predicate)
        except Linq.LinqException:
            return None

    def to_list(self):
        """
        Converts LinqIterable to list
        :returns: list
        :rtype: list
        """
        return list(self.iterable)

    def to_dictionary(self, key_selector=None, value_selector=None, unique=True):
        """
        Converts LinqIterable to dictionary
        :param key_selector: function which takes item and returns key for it
        :param value_selector: function which takes item and returns value for it
        :param unique: boolean, if True that will throw exception if keys are not unique
        :returns: dict
        :rtype: dict
        """
        result = {}
        keys = set() if unique else None

        for i in self.iterable:
            key = key_selector(i) if key_selector is not None else i
            value = value_selector(i) if value_selector is not None else i
            if unique:
                if key in keys:
                    raise Linq.LinqException("Key '" + repr(key) + "' is used more than once.")
                keys.add(key)
            result[key] = value
        return result

    def where(self, predicate):
        """
        Returns items which matching predicate function
        :param predicate: Function which takes item as argument and returns bool
        :returns: results wrapped with Linq
        :rtype: Linq
        """
        
        result = []
        for i in self.iterable:
             
            try:
                if predicate(i):
                    result.append(i)            
            except Exception:
                pass

        return Linq(result)

    def distinct(self, key_selector=None):
        """
        Filters distinct values from enumerable
        :param key_selector: function which takes item and returns key for it
        :returns: results wrapped with Linq
        :rtype: Linq
        """
        key_selector = key_selector if key_selector is not None else lambda item: item
        keys = set()
        return Linq([i for i in self.iterable if key_selector(i) not in keys and not keys.add(key_selector(i))])

    def group_by(self, key_selector=None, value_selector=None):
        """
        Groups given items by keys.
        :param key_selector: function which takes item and returns key for it
        :param value_selector: function which takes item and returns value for it
        :returns: Dictionary, where value if Linq for given key
        :rtype: dict
        """
        key_selector = key_selector if key_selector is not None else lambda item: item
        value_selector = value_selector if value_selector is not None else lambda item: item

        result = {}
        for i in self.iterable:
            key = key_selector(i)
            if result.__contains__(key):
                result[key].append(value_selector(i))
            else:
                result[key] = [value_selector(i)]
        for key in result:
            result[key] = Linq(result[key])
        return result

    def order_by(self, value_selector=None, comparer=None, descending=False):
        """
        Orders items.
        :param value_selector: function which takes item and returns value for it
        :param comparer: function which takes to items and compare them returning int
        :param descending: shows how items will be sorted
        """
        return Linq(sorted(self.iterable, comparer, value_selector, descending))

    def take(self, number):
        """
        Takes only given number of items, of all available items if their count is less than number
        :param number: number of items to get
        :returns: results wrapped with Linq
        :rtype: Linq
        """
        def internal_take(iterable, number):
            count = 0
            for i in iterable:
                count += 1
                if count > number:
                    break
                yield i

        return Linq(internal_take(self.iterable, number))

    def skip(self, number):
        """
        Skips given number of items in enumerable
        :param number: number of items to get
        :returns: results wrapped with Linq
        :rtype: Linq
        """
        def internal_skip(iterable, number):
            count = 0
            for i in iterable:
                count += 1
                if count <= number:
                    continue
                yield i

        return Linq(internal_skip(self.iterable, number))

    def select(self, selector):
        """
        Converts items in list with given function
        :param selector: Function which takes item and returns other item
        :returns: results wrapped with Linq
        :rtype: Linq
        """
        return Linq([selector(i) for i in self.iterable])

    def select_many(self, selector):
        """
        Converts items in list with given function
        :param selector: Function which takes item and returns iterable
        :returns: results wrapped with Linq
        :rtype: Linq
        """
        return Linq([i for i in [selector(sub) for sub in self.iterable]])

    def foreach(self, func):
        """
        Allows to perform some action for each object in iterable, but not allows to redefine items
        :param func: Function which takes item as argument
        :returns: self
        :rtype: Linq
        """
        for i in self.iterable:
            func(i)
        return self

    def concat(self, iterable):
        """
        Concats two iterables
        :param iterable: Any iterable
        :returns: self
        :rtype: Linq
        """
        return Linq(itertools.chain(self.iterable, iterable))

    def concat_item(self, item):
        """
        Concats iterable with single item
        :param item: Any item
        :returns: self
        :rtype: Linq
        """
        return Linq(itertools.chain(self.iterable, [item]))

    def except_for(self, iterable):
        """
        Filters items except given iterable
        :param iterable: Any iterable
        :returns: self
        :rtype: Linq
        """
        return Linq([i for i in self.iterable if i not in iterable])

    def intersect(self, iterable):
        """
        Intersection between two iterables
        :param iterable: Any iterable
        :returns: self
        :rtype: Linq
        """
        return Linq([i for i in self.iterable if i in iterable])


class LinqException(Exception):
    """
    Special exception to be thrown by Linq
    """
    pass



# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import pathlib
from lemondb.plugin.base import BasePlugin
import json
from lemondb.types import (
    Optional
)

class LemonPlugin(BasePlugin):
    """
    The base plugin for LemonDB. This create a standard
    JSON db.
    """

    def __init__(self, **options):
        self.options = options
        super(LemonPlugin, self).__init__(**self.options)

    def _init_db(self):
        """
        Initialize the db if not exist.
        """

        path = pathlib.Path(self.name)
        with open(path.absolute(), 'w') as f:
            json.dump({self.kwargs.get('table_name'): {}}, f)
        

    def run(
        self, 
        name=None, 
        document_cls=None, 
        plugin_cls=None, 
        middleware_cls=None, 
        **kwargs
    ):

        self.name = name
        self.document_cls = document_cls
        self.plugin_cls = plugin_cls
        self.middleware_cls = middleware_cls
        self.kwargs = kwargs
        



# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

try:
    from sidle.enc import SidleEncryption
    import sidle.errors as error
except ModuleNotFoundError:
    SidleEncryption = None


from lemondb.middleware import BaseMiddleware
from lemondb.types import (
    Optional,
    Mapping,
    Dict,
    Any
)
import pathlib
from lemondb.serializer import JSONSerializer

class SidleMiddleware(BaseMiddleware):
    """
    A encryption middleware for `SidleEncryption`.
    That can encrypted with password
    """

    serializer: SidleEncryption
    def __init__(self, password: str, jsonserializer: Optional[str] = 'json'):
        """
        Initialize SidleMiddleware.

        Parameter:
            password (str):
                The password key for the database.
        """

        if not SidleEncryption:
            raise ModuleNotFoundError('Sidle is not installed. `pip install sidle` might solved this')

        self.password = password
        self.serializer = SidleEncryption(self.password)
        self.json = JSONSerializer(jsonserializer)

    def read(self, path: str) -> Dict[str, Any]:
        if isinstance(path, pathlib.Path):
            path = path.absolute()
        else:
            path = path


        with open(path, 'rb') as f:
            raw = self.serializer.decrypt(
                f.read()
            )
            if not raw:
                raise error.PasswordError("\"%s\" is not a valid password" % (self.password))

            if isinstance(raw, bytes):
                raw = raw.decode()
            else:
                raw = str(raw)
                
            data = self.json.loads(
                s=raw
            )
        
        return data

    def write(
        self, 
        item: Mapping, 
        path: str, 
        mode: Optional[str] = 'rb+'
    ):
        
        """
        Write the given item with encryption
        """
        if mode == 'w':
            mode = 'wb'
        elif mode == 'w+':
            mode == 'wb+'
            
        if isinstance(item, dict):
            item = self.json.dumps(item)
        else:
            item = item

        item = self.serializer.encrypt(item)

        if isinstance(path, pathlib.Path):
            path = path.absolute()
        else:
            path = path
            
        with open(path, mode) as f:
            f.write(item)

    def delete(
        self, 
        key: str, 
        path: str, 
        all: Optional[bool]
    ):
        """
        Delete the given key and write to the given
        file path
        """

        data = self.read(path)
        
        #: TODO: Kindly remember that we convert
        #: the data.items into a list to avoid
        #: RuntimeError: changed sized during iteration
        #: This happens because it return a iterator
        #: instead of a list.

        for table, value in list(data.items()):
            for k,v in list(value.items()):
                if isinstance(key, (tuple, list)):
                    if v in key:
                        del data[table][k]
                        
                    

                if key == v and all:
                    del data[table][k]
                    
                elif key == v and not all:
                    del data[table][k]
                    break
        
        
        self.write(data, path, mode='wb')   


# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from lemondb.types import (
    Optional,
    Mapping
)
from sidle.enc import SidleEncryption
import socket
import json
import sys

class BaseServer:
    """
    Base class for the server.
    """

    def create(self):
        """Base Server for creating server. Notimplemented"""
        raise NotImplementedError
        
    def run(self):
        """Run the server given the host."""
        raise NotImplementedError

class BaseClient:
    """
    Base class for handling client connection
    """
    
    def connect(self):
        raise NotImplementedError

    def send(self, data: Mapping):
        raise NotImplementedError
        

class LemonServer(BaseServer):
    """
    LemonServer is a socket server that used to store LemonDB over
    the socket connection. The database is stored on the lemondb environment
    (~/.lemondb). It can do all operation and the packets / data sent from
    the client should be encrypted using SidleEncryption. See LemonClient
    for the example client. 

    Parameters:
        
        host (str):
            The host of the server to be bind to. The host pattern should
            be 0.0.0.0:3000 where 3000 is the port and 0.0.0.0 is the host
            ofcourse.

        key (Optional[str]):
            The password key for the server. Optional. Set it to None if 
            you don't want to add password key.

    Example (Cannot be used directly):

        >>> server = LemonServer('0.0.0.0:3000', 'simple_password', ...)
        >>> server.run()

    """
    
    #: Initialize socket object.
    socket: socket.socket
    server_loop: bool
    
    #: Initialize Sidle encryption object.
    enc: SidleEncryption
    enc_key: str = 'lemondb.server/0.1'

    def __init__(
        self, 
        host: str, 
        key: Optional[str] = None, 
        **options
    ):
        """
        Parameters:
        
            host (str):
                The host of the server to be bind to. The host pattern should
                be 0.0.0.0:3000 where 3000 is the port and 0.0.0.0 is the host
                ofcourse.

            key (Optional[str]):
                The password key for the server. Optional. Set it to None if 
                you don't want to add password key.

            Keyword Arguments (kwargs):
            
                reuse_socket (Optional[bool]: True):
                    Allow the socket to be reused.
                
                client (Optional[int]: 1):
                    Set how many client the socket should listen to.

                buffer_value (Optional[int]: 65537):
                    The receive buffer value.

        """

        self.host = host
        self.key  = key
        self.opt  = options
        self.enc  = SidleEncryption(self.enc_key)
        self.buff = self.opt.get('buffer_value', 65537)

        if isinstance(host, str):
            self.host = self.__parse_host(host)


    def create(self) -> socket.socket:
        """
        Create the socket server given the host and configuration.
        """
        #: Get all kwargs option 
        client = self.opt.get('client', 1)
        reuse_socket = self.opt.get('reuse_socket', True)

        self.socket = socket.socket(
            family=socket.AF_INET, 
            type=socket.SOCK_STREAM
        )
        if reuse_socket:
            
            self.socket.setsockopt(
                socket.SOL_SOCKET, 
                socket.SO_REUSEADDR, 
                1
            )
        
        self.socket.bind(self.host)
        self.socket.listen(client)
        self.server_loop = True

        return self.socket

    def run(self, dbclass):
        server = self.create()
        try:
            while self.server_loop:
                self._handle_request(server, dbclass)

        except KeyboardInterrupt:
            sys.exit()

    def _handle_request(self, server: socket.socket, dbclass):
        """
        Handle all request sent to the server.
        """
        connection, address = server.accept()
        raw_data = connection.recv(self.buff)
        
        #: The sent data should be json formatted.
        if raw_data:
            decrypted = json.loads(
                self.enc.decrypt(raw_data)
            )

            op = decrypted.get('op', None)
            data = decrypted.get('data', None)
        
        try:
            operation = dbclass.__getattribute__(op)
            operation(data)
        except AttributeError:
            dbclass.logger.error('"%s" is not an operation' % (op))


    def __parse_host(self, host: str):
        lst = host.split(':')
        try:
            _host, port = lst
        except Exception:
            raise
        
        return (_host, int(port))

class LemonClient(BaseClient):
    """
    LemonClient is a client handler for the server.
    It handle the operation of sending data and connection
    via LemonServer. Also it has an encryption for the LemonServer
    which is lemondb.server/{version} 
    """

    #: Initialize Socket object
    _socket: socket.socket

    #: Initialize Sidle encryption object.
    enc: SidleEncryption
    enc_key: str = 'lemondb.server/0.1'

    def __init__(
        self, 
        host: str, 
        key: Optional[str] = None,
        **option
    ):

        self.host = host
        self.key  = key
        self.opt  = option

        if isinstance(host, str):
            self.host = self.__parse_host(host)

        if option.get('auto_connect', True):
            self.connect()


    def connect(self):
        """
        Create a connection between the socket server that
        handles data and inorder to send data.
        """

        self.__initialize_socket()
        try:
            self._socket.connect(self.host)
        except Exception:
            return False


        return True
        

    def send(self, op: str, data: Mapping):
        """
        Send raw data given the operation to the
        socket server.
        """

        all_op = ['insert', 'update', 'delete', 'search']
        if op not in all_op:
            raise TypeError('%s is not an operation' % (op))

        raw = json.dumps({
            'op': op,
            'data': data
        })

        self._socket.send(
            self.enc.encrypt(raw)
        )

        return raw


    def __initialize_socket(self):
        """
        Initialize socket object
        """
        self._socket = socket.socket(
            family=socket.AF_INET, 
            type=socket.SOCK_STREAM
        )
        return self._socket

    def __parse_host(self, host: str):
        lst = host.split(':')
        try:
            _host, port = lst
        except Exception:
            raise
        
        return (_host, int(port))

    



# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from setuptools import setup, find_packages
from lemondb import __version__


def get_long_description():
    with open("README.md", encoding="utf-8") as f:
        readme = f.read()

    return readme

setup(
    
    name="lemondb", 
    description="Simple and lightweight document oriented database",
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    author = 'Zenqi',
    license = 'MIT',
    version = __version__,
    packages = [p for p in find_packages() if 'test' not in p],
    install_requires=[
        'sidle>=0.0.2',
        'loguru>=0.5.3'
    ],

)



# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import pathlib
from lemondb.middleware.base import BaseMiddleware
from lemondb.serializer import JSONSerializer
from lemondb.types import (
    Optional,
    Mapping,
    Dict,
    Any
)
from lemondb.utils import iterate_dict

class JsonMiddleware(BaseMiddleware):
    """
    A main class for handling JSON data inorder to read
    and write and delete item.
    """
    
    serializer: JSONSerializer

    def __init__(self, serializer: Optional[str] = 'json'):
        self.serializer = JSONSerializer(serializer)
        super().__init__()
        
    def read(self, path: str) -> Dict[str, Any]:
        """
        Read the file and return the data

        Parameter:
            path (str):
                The path for the file to be red. It can 
                be pathlib.Path instance or a str.
        """

        if isinstance(path, pathlib.Path):
            path = path.absolute()
        else:
            path = path

        with open(path, 'r+') as f:
            return self.serializer.load(f)

    def write(
        self, 
        item: Mapping, 
        path: str, 
        mode: Optional[str] = 'r+'
    ):
        
        """
        Write the given data to the file
        """
        if isinstance(path, pathlib.Path):
            path = path.absolute()
        else:
            path = path

        if not path.exists():
            raise FileNotFoundError('"%s" not found' % (path.absolute()))

        data = self.read(path); data.update(item)
        with open(path, mode) as f:
            self.serializer.dump(data, f)

            
    def delete(
        self, 
        key: Mapping, 
        path: str, 
        all: Optional[bool] = True
    ):
        
        """
        Delete the given key and write to the given
        file path
        """

        data = self.read(path)
        
        #: TODO: Kindly remember that we convert
        #: the data.items into a list to avoid
        #: RuntimeError: changed sized during iteration
        #: This happens because it return a iterator
        #: instead of a list.

        for table, value in list(data.items()):
            for k,v in list(value.items()):
                if isinstance(key, (tuple, list)):
                    if v in key:
                        del data[table][k]
                        
                    

                if key == v and all:
                    del data[table][k]
                    
                elif key == v and not all:
                    del data[table][k]
                    break
        
        
        self.write(data, path, mode='w')


# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import importlib
from lemondb.serializer.base import BaseSerializer
from lemondb.types import Optional


class JSONSerializer(BaseSerializer):
    """
    A simple JSON Serializer class for handling 
    dump and loading json from different JSON serializer
    like `simplejson`, `rapidjson`, `ujson` without changing
    the standard json format.

    Available serializer:
        `json`
        `simplejson`
        `ujson`
        `hyperjson`
        `rapidjson`
    
    Example:
        >>> from lemondb.serializer import Serializer
        >>> json = Serializer(name='json') #: standard json serializer
        >>> json.dumps(...)
        >>> ...
        >>> simplejson = Serializer(name='simplejson')
        >>> simplejson.dumps(...)
        >>> ...
    
    """

    _available_serializer = [
        'json', 'ujson', 'simplejson', 'rapidjson',
        'hyperjson'
    ]

    def __init__(self, name: Optional[str] = 'json'):
        """
        Initialize JSON Serializer
        """

        if name not in self._available_serializer:
            raise ValueError('"%s" is not found' % (name))

        try:
            self.serializer = importlib.import_module(name)
        except ImportError:
            self.serializer = importlib.import_module('json')

    def dump(self, *args, **kwargs):
        """
        Dump the JSON Object onto a File IO
        """
        return self.serializer.dump(*args, **kwargs)

    def dumps(self, *args, **kwargs):
        """
        Dump the JSON Object to a string
        """
        return self.serializer.dumps(*args, **kwargs)

    def load(self, *args, **kwargs):
        """
        Load the JSON Object from the File IO given
        """
        return self.serializer.load(*args, **kwargs)

    def loads(self, *args, **kwargs):
        """
        Load the JSON Object from the string given
        """
        return self.serializer.loads(*args, **kwargs)




# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import Mapping

def iterate_dict(item: Mapping):
    """
    Iterate through all nested dictionaries
    and yield them.

    Parameters:
        item (Mapping):
            The item /dict to be iterate

    """
    for k,v in item.items():
        if isinstance(v, Mapping):
            yield from iterate_dict(v)
        else:
            yield k,v 

def deco(dec, condition):
    """
    Set the function a decorator if the
    condition is True else ignore.

    Parameters:
        decorator (Function):
            A decorator function
        
        condition (bool):
            The codition inorder to set the decorator.

    """

    def wrapper(func):
        if not condition:
            #: Ignore the decorator.
            return func

        return dec(func)

    return wrapper





# Copyright (c) 2021 Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from lemondb.types import (
    Optional,
    Any,
    Mapping,
    Middleware,
    List,
    Dict
)
from lemondb.middleware import JsonMiddleware


class Document:
    """
    A document class used for handling file operations
    such as writting, reading and deleting item.
    """

    def __init__(
        self,
        path: str,
        middleware_cls: Middleware = JsonMiddleware #: The middleware for the document. 
        # Default: JsonMiddleware

    ):

        self.path = path
        try:
            self.middleware = middleware_cls()
        except TypeError:
            # TODO: Middleware class already initialized.
            self.middleware = middleware_cls
    
    def __getattr__(self, item: str):
        return self.__getitem__(
            key=item
        )

    def __getitem__(
        self, 
        key: str, 
        default: Optional[Any] = None
    ):

        pass

    def read(self) -> dict:
        return self.middleware.read(path=self.path)

    def write(
        self, 
        item: Mapping, 
        raw: Optional[bool] = False,
        mode: Optional[str] = 'r+'
    ):
        """
        Write the item to the document
        """

        data = self.read()
        if not raw:
            item = self._increment(
                data=data,
                item=item
            )
        else:
            item = item

        
        return self.middleware.write(item, path=self.path, mode=mode)


    def delete(self, item: Mapping, all: Optional[bool] = True):
        return self.middleware.delete(item, path=self.path, all=all)
    
    def _increment(self, data: Mapping, item: Mapping) -> Mapping:
        """
        Increment the given item to the last given data.
        """

        raw = []
        for k, v in data.items():
            if v:
                raw = list(v.items())
            else:
                v.update({'0': item}); return data


        k,v = raw[-1]
        i = {int(k) + 1: item}
        for k,v in data.items(): v.update(i)

        return data
